#!/usr/bin/env python3

import copy
import os
import sys
from io import TextIOWrapper

from .args import args
from .atoms import AtomInfo
from .coefficient import get_coefficient
from .data import Data_All_MO, Data_MO
from .eigenvalues import Eigenvalues, get_eigenvalues
from .functions_info import get_functions_info
from .utils import debug_print, space_separated_parsing


def is_this_row_for_coefficients(words: "list[str]") -> bool:
    # min: 4 coefficients and other words => 5 words
    if 5 <= len(words) <= 9 and words[0].isdigit():
        return True
    else:
        return False


def need_to_skip_this_line(words: "list[str]") -> bool:
    if len(words) <= 1:
        return True
    else:
        return False


def need_to_create_results_for_current_mo(words: "list[str]", is_reading_coefficients: bool) -> bool:
    if is_reading_coefficients and len(words) <= 1:
        return True
    else:
        return False


def need_to_get_mo_sym_type(words: "list[str]", start_mo_coefficients: bool) -> bool:
    if not start_mo_coefficients and len(words) == 3 and words[0] == "Fermion" and words[1] == "ircop":
        return True
    return False


def need_to_start_mo_section(words: "list[str]", start_mo_coefficients: bool) -> bool:
    if not start_mo_coefficients and words[1] == "Electronic" and words[2] == "eigenvalue" and "no." in words[3]:
        return True
    elif not start_mo_coefficients and words[1] == "Positronic" and words[2] == "eigenvalue" and "no." in words[3]:
        return True
    return False


def get_dirac_filename() -> str:
    if not args.file:
        sys.exit("ERROR: DIRAC output file is not given. Please use -f option.")
    return args.file


def check_start_vector_print(words: "list[str]") -> bool:
    # ****************************** Vector print ******************************
    if len(words) < 4:
        return False
    elif words[1] == "Vector" and words[2] == "print":
        return True
    return False


def check_end_vector_print(
    words: "list[str]",
    start_vector_print: bool,
    start_mo_section: bool,
    start_mo_coefficients: bool,
    is_reading_coefficients: bool,
) -> bool:
    # https://github.com/kohei-noda-qcrg/summarize_dirac_dfcoef_coefficients/issues/7#issuecomment-1377969626
    if len(words) >= 2 and start_vector_print and start_mo_section and not start_mo_coefficients and not is_reading_coefficients:
        return True
    return False


def get_output_path() -> str:
    if args.output is None:
        output_name = "sum_dirac_dfcoef.out"
        output_path = os.path.join(os.getcwd(), output_name)
    else:
        output_name = args.output
        output_path = os.path.abspath(output_name)
    return output_path


def should_write_positronic_results_to_file() -> bool:
    if args.all_write or args.positronic_write:
        return True
    else:
        return False


def should_write_electronic_results_to_file() -> bool:
    if args.all_write or not args.positronic_write:
        return True
    else:
        return False


def write_results(file: TextIOWrapper, data_all_mo: "list[Data_MO]") -> None:
    """
    Write results to stdout
    """

    for mo in data_all_mo:
        digit_int = len(str(int(mo.mo_energy)))  # number of digits of integer part
        # File write but if args.compress is True \n is not added
        mo_info_energy = f"{mo.mo_info} {mo.mo_energy:{digit_int}.{args.decimal}f}" + ("\n" if not args.compress else "")
        file.write(mo_info_energy)

        for c in mo.coef_list:
            for idx in range(c.multiplication):
                percentage = c.coefficient / mo.norm_const_sum * 100
                label = f"{c.function_label}({c.start_idx + idx})" if c.need_identifier else c.function_label
                output_str: str
                if args.compress:
                    output_str = f" {label} {percentage:.{args.decimal}f}"
                else:
                    output_str = f"{label:<12} {percentage:{args.decimal+4}.{args.decimal}f} %\n"
                file.write(output_str)
        file.write("\n")  # add empty line
        debug_print(f"sum of coefficient {mo.norm_const_sum:.{args.decimal}f}")


def main() -> None:
    start_mo_coefficients: bool = False
    start_mo_section: bool = False
    is_reading_coefficients: bool = False
    start_vector_print: bool = False
    is_electronic: bool = False
    mo_sym_type: str = ""

    dirac_filename: str = get_dirac_filename()
    dirac_output = open(dirac_filename, encoding="utf-8")
    dirac_output.seek(0)  # rewind to the beginning of the file
    functions_info = get_functions_info(dirac_output)
    eigenvalues: Eigenvalues = get_eigenvalues(dirac_output)
    with open(get_output_path(), "w", encoding="utf-8") as file:
        line = ""
        for symmetry_type, d in eigenvalues.items():
            line += f"{symmetry_type} "
            for eigenvalue_type, num in d.items():
                line += f"{eigenvalue_type} {num} "
        line += "\n"
        file.write(line)

    data_mo = Data_MO()
    data_all_mo = Data_All_MO()
    used_atom_info: dict[str, AtomInfo] = dict()
    current_atom_info = AtomInfo()
    dirac_output.seek(0)  # rewind to the beginning of the file
    for line_str in dirac_output:
        words: "list[str]" = space_separated_parsing(line_str)

        if not start_vector_print:
            if check_start_vector_print(words):
                start_vector_print = True
            continue

        if need_to_get_mo_sym_type(words, start_mo_coefficients):
            mo_sym_type = words[2]

        elif need_to_skip_this_line(words):
            if need_to_create_results_for_current_mo(words, is_reading_coefficients):
                start_mo_coefficients = False
                data_mo.fileter_coefficients_by_threshold()
                if is_electronic:
                    data_all_mo.electronic.append(copy.deepcopy(data_mo))
                else:  # Positronic
                    data_all_mo.positronic.append(copy.deepcopy(data_mo))
                debug_print(f"End of reading {data_mo.electron_num}th MO")
                is_reading_coefficients = False

        elif need_to_start_mo_section(words, start_mo_coefficients):
            """
            (e.g.)
            words = ["*", "Electronic", "eigenvalue", "no.", "22:", "-2.8417809384721"]
            words = ["*", "Electronic", "eigenvalue", "no.122:", "-2.8417809384721"]
            """
            start_mo_section = True
            start_mo_coefficients = True
            if words[1] == "Positronic":
                is_electronic = False
            elif words[1] == "Electronic":
                is_electronic = True
            else:
                raise Exception(f"Unknown MO type, MO_Type={words[1]}")
            try:
                electron_num = int(words[-2][:-1].replace("no.", ""))
            except ValueError:
                # If *** is printed, we have no information about what number this MO is.
                # Therefore, we assume that electron_num is the next number after prev_electron_num.
                prev_electron_num = data_mo.electron_num  # prev_electron is the number of electrons of the previous MO
                electron_num = prev_electron_num + 1
            mo_energy = float(words[-1])
            mo_info = (
                f"{mo_sym_type} {electron_num}"
                if args.compress
                else (f"Electronic no. {electron_num} {mo_sym_type}" if is_electronic else f"Positronic no. {electron_num} {mo_sym_type}")
            )
            # Here is the start point of reading coefficients of the current MO
            data_mo.reset()  # reset data_mo because we need to delete data_mo of the previous MO
            data_mo.electron_num = electron_num
            data_mo.mo_energy = mo_energy
            data_mo.mo_info = mo_info
            used_atom_info.clear()  # reset used_atom_info because we need to delete used_atom_info of the previous MO

        elif check_end_vector_print(
            words,
            start_vector_print,
            start_mo_section,
            start_mo_coefficients,
            is_reading_coefficients,
        ):
            # End of reading coefficients
            break

        # Read coefficients or the end of coefficients section
        elif start_mo_coefficients:
            if not is_this_row_for_coefficients(words):
                continue
            is_reading_coefficients = True
            component_func = "large" if line_str[10] == "L" else ("small" if line_str[10] == "S" else "")  # CLS
            symmetry_label = line_str[12:15].strip()  # REP (e.g. "Ag "), symmetry_label="Ag"
            atom_label = line_str[15:18].strip()  # NAMN (e.g. "Cm "), atom_labe="Cm"
            gto_type = line_str[18:22].strip()  # GTOTYP (e.g. "s   "), gto_type="s"
            label = symmetry_label + atom_label

            if current_atom_info.count_remaining_functions() == 0 or label != current_atom_info.label:
                # First, we need to read information about the current atom.
                if label not in used_atom_info:
                    # It is the first time to read information about the current atom.
                    cur_atom_start_idx = 1
                else:
                    # It is not the first time to read information about the current atom.
                    # So we need to read information about the previous atom from used_atom_info.
                    # current start_idx = previous start_idx + previous mul
                    cur_atom_start_idx = used_atom_info[label].start_idx + used_atom_info[label].mul
                # Validate start_idx
                if cur_atom_start_idx not in functions_info[component_func][symmetry_label][atom_label]:
                    raise Exception(f"start_idx={cur_atom_start_idx} is not found in functions_info[{component_func}][{symmetry_label}][{atom_label}]")
                # We can get information about the current atom from functions_info with start_idx.
                current_atom_info = copy.deepcopy(functions_info[component_func][symmetry_label][atom_label][cur_atom_start_idx])
                # Update used_atom_info with current_atom_info
                used_atom_info[label] = copy.deepcopy(current_atom_info)

            current_atom_info.decrement_function(gto_type)
            data_mo.add_coefficient(get_coefficient(line_str, functions_info, current_atom_info.start_idx))

    # End of reading file
    # Write results to the file
    file = open(get_output_path(), "a", encoding="utf-8")
    if not args.no_sort:
        data_all_mo.electronic.sort(key=lambda x: x.mo_energy)
        data_all_mo.positronic.sort(key=lambda x: x.mo_energy)
    if should_write_positronic_results_to_file():  # Positronic
        # Write positronic results to the file
        write_results(file, data_all_mo.positronic)
        file.write("\n")  # Add a blank line
    if should_write_electronic_results_to_file():  # Electronic
        # Write electronic results to the file
        write_results(file, data_all_mo.electronic)
    file.close()



import argparse


class PrintVersionExitAction(argparse.Action):
    def __init__(self, option_strings, dest=argparse.SUPPRESS, default=argparse.SUPPRESS, help=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        from .__about__ import __version__

        print(f"{__version__}")
        exit()


def parse_args() -> "argparse.Namespace":
    parser = argparse.ArgumentParser(
        description="Summarize the coefficients from DIRAC output file that *PRIVEC option is used. (c.f. http://www.diracprogram.org/doc/master/manual/analyze/privec.html)"
    )
    parser.add_argument("-i", "--input", type=str, required=True, help="(required) file name of DIRAC output", dest="file")
    parser.add_argument("-o", "--output", type=str, help="Output file name. Default: sum_dirac_dfcoef.out", dest="output")
    parser.add_argument(
        "-c",
        "--compress",
        action="store_true",
        help="Compress output. Display all coefficients on one line for each MO. This options is useful when you want to use the result in a spreadsheet like Microsoft Excel.",
        dest="compress",
    )
    parser.add_argument(
        "-t", "--threshold", type=float, default=0.1, help="threshold. Default: 0.1 %% (e.g) --threshold=0.1 => print orbital with more than 0.1 %% contribution", dest="threshold"
    )
    parser.add_argument(
        "-d",
        "--decimal",
        type=int,
        default=5,
        choices=range(1, 16),
        help="Set the decimal places. Default: 5 (e.g) --decimal=3 => print orbital with 3 decimal places (0.123, 2.456, ...). range: 1-15",
        dest="decimal",
    )
    parser.add_argument("-a", "--all-write", action="store_true", help="Print all MOs(Positronic and Electronic).", dest="all_write")
    parser.add_argument("-p", "--positronic-write", action="store_true", help="Print only Positronic MOs.", dest="positronic_write")
    parser.add_argument("-v", "--version", action=PrintVersionExitAction, help="Print version and exit", dest="version")
    parser.add_argument("--debug", action="store_true", help="print debug output (Normalization constant, Sum of MO coefficient)", dest="debug")
    parser.add_argument("--no-sort", action="store_true", help="Don't sort the output by MO energy")
    # If -v or --version option is used, print version and exit
    return parser.parse_args()


args = parse_args()



def space_separated_parsing(line: str) -> "list[str]":
    import re

    words = re.split(" +", line.rstrip("\n"))
    return [word for word in words if word != ""]


def debug_print(message: str):
    from .args import args

    # print debug message if --debug option is used
    if args.debug:
        print(message)

def is_float(parameter: str):
    if not parameter.isdecimal():
        try:
            float(parameter)
            return True
        except ValueError:
            return False
    else:
        return False


import re
from io import TextIOWrapper
from typing import Dict

from .utils import space_separated_parsing


# type definition eigenvalues
# type eigenvalues = {
#     "E1g": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
#     "E1u": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
# }
class Eigenvalues(Dict[str, Dict[str, int]]):
    pass


def get_eigenvalues(dirac_output: TextIOWrapper) -> Eigenvalues:
    def is_end_of_read(line) -> bool:
        if "Occupation" in line or "HOMO - LUMO" in line:
            return True
        return False

    def is_eigenvalue_type_written(words: "list[str]") -> bool:
        # closed shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1043
        # open shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1053
        # virtual eigenvalues: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1064
        if "*" == words[0] and "Closed" == words[1] and "shell," == words[2]:
            return True
        elif "*" == words[0] and "Open" == words[1] and "shell" == words[2]:
            return True
        elif "*" == words[0] and "Virtual" == words[1] and "eigenvalues," == words[2]:
            return True
        return False

    def get_current_eigenvalue_type(words: "list[str]") -> str:
        # words[0] = '*', words[1] = "Closed" or "Open" or "Virtual"
        current_eigenvalue_type = words[1].lower()
        return current_eigenvalue_type

    def get_symmetry_type_standard(words: "list[str]") -> str:
        current_symmetry_type = words[3]
        return current_symmetry_type

    def get_symmetry_type_supersym(words: "list[str]") -> str:
        # https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1097-1105
        # FORMAT '(/A,I4,4A,I2,...)'
        # DATA "* Block",ISUB,' in ',FREP(IFSYM),":  ",...
        # ISUB might be **** if ISUB > 9999 or ISUB < -999 because of the format
        # Therefore, find 'in' word list and get FREP(IFSYM) from the word list
        # FREP(IFSYM) is a symmetry type
        idx = words.index("in")
        current_symmetry_type = words[idx + 1][: len(words[idx + 1]) - 1]
        return current_symmetry_type

    scf_cycle = False
    eigenvalues_header = False
    print_type = ""  # 'standard' or 'supersymmetry'
    eigenvalues = Eigenvalues()
    current_eigenvalue_type = ""  # 'closed' or 'open' or 'virtual'
    current_symmetry_type = ""  # 'E1g' or 'E1u' or "E1" ...

    for line in dirac_output:
        words: "list[str]" = space_separated_parsing(line)

        if len(words) == 0:
            continue

        if "SCF - CYCLE" in line:
            scf_cycle = True
            continue

        if scf_cycle and not eigenvalues_header:
            if "Eigenvalues" == words[0]:
                eigenvalues_header = True
            continue

        if print_type == "":  # search print type (standard or supersymmetry)
            if "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
                print_type = "standard"
                current_symmetry_type = get_symmetry_type_standard(words)
                eigenvalues.setdefault(current_symmetry_type, {'closed': 0, 'open': 0, 'virtual': 0})
            elif "* Block" in line:
                print_type = "supersymmetry"
                current_symmetry_type = get_symmetry_type_supersym(words)
                eigenvalues.setdefault(current_symmetry_type, {'closed': 0, 'open': 0, 'virtual': 0})
            continue

        if print_type == "standard" and "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
            current_symmetry_type = get_symmetry_type_standard(words)
            eigenvalues.setdefault(current_symmetry_type, {'closed': 0, 'open': 0, 'virtual': 0})
        elif print_type == "supersymmetry" and "* Block" in line:
            current_symmetry_type = get_symmetry_type_supersym(words)
            eigenvalues.setdefault(current_symmetry_type, {'closed': 0, 'open': 0, 'virtual': 0})
        elif is_eigenvalue_type_written(words):
            current_eigenvalue_type = get_current_eigenvalue_type(words)
        elif is_end_of_read(line):
            break
        else:
            start_idx = 0
            while True:
                # e.g. -775.202926514  ( 2) => 2
                regex = r"\([ ]*[0-9]+\)"
                match = re.search(regex, line[start_idx:])
                if match is None:
                    break
                num = int(match.group()[1 : len(match.group()) - 1])
                eigenvalues[current_symmetry_type][current_eigenvalue_type] += num
                start_idx += match.end()

    print(f"eigenvalues: {eigenvalues}")
    return eigenvalues